<?php
namespace SupplyTek\Utilities;

use Doctrine\ORM\EntityManager;
use Doctrine\DBAL\Connection;
use Zeedhi\Framework\DataSource\Manager\Doctrine\NameProvider;
use Zeedhi\Framework\DataSource\FilterCriteria;
use Zeedhi\Framework\DataSource\Configuration;
use Zeedhi\Framework\DTO\Request;

/**
 * General class that perform queries on database
 * 
 * This class has functions that process FilterCriteria, build and perform 
 * queries, and format the result.
 * 
 * @author Danilo D. Vaz
 */
class CrudSupplyTek
{
    protected $connection;
    protected $nameProvider;
    protected $productPath;

    /**
     * Class constructor
     * 
     * Initialize an object's properties upon creation of the object.
     * 
     * @param EntityManager $entityManager Object that allows to manipulate DataBase connection
     * @param NameProvider $nameProvider Object that allows to manipulate datasources
     * @param string $productPath String with product path
     * 
     * @return void
     * 
     * @access public
     * @author Danilo D. Vaz
     */
    public function __construct (EntityManager $entityManager, NameProvider $nameProvider, string $productPath) 
    {
        $this->connection = $entityManager->getConnection();
        $this->nameProvider = $nameProvider;
        $this->productPath = $productPath;
    }

    /**
     * Main function that process request and provide data
     * 
     * Receives the request sent from the frontend and additional filters,
     * get the query from datasource file, process filters, build and execute
     * the query. Provides the result formated.
     * 
     * @param Request $request Object that receives data from the caller
     * @param array $addtionalFilter Associative array with filter ['COLUMN'=>'VALUE']
     *
     * @return array
     * 
     * @access public
     * @author Danilo D. Vaz
     */
    public function find (Request $request, array $addtionalFilter = array())
    {
        $filterUri = $request->getParameters();
        $filterCriteria = $request->getFilterCriteria();

        $dataSource = $this->getDataSource($filterCriteria);
        $filter = $this->processFilter($filterCriteria, $filterUri, $addtionalFilter);
        $query = $this->buildQuery($filterCriteria, $dataSource, $filter);
        $result = $this->executeQuery('', $filter, $query, $dataSource->getConditions());
        $this->buildUri($result, array(), $filterUri, $dataSource);
        
        if (count($filterUri) && count($result)) {
            $result = reset($result);
        }

        $data = $this->buildData($result, $filterCriteria);

        return $data;
    }

    /**
     * Function that performs query on database
     * 
     * Can receives the datasource name or the query. If informed datasoure name,
     * query and column types will be geted in datasource. Returns the queries
     * result.
     * 
     * @param string $dataSourceName String that receives datasource name
     * @param array $filter Associative array with filter ['COLUMN'=>'VALUE']
     * @param string $query String that receives the query
     * @param array $columnTypes Array that receives the columns types ['COLUMN'=>'TYPE']
     *
     * @return array
     * 
     * @access public
     * @author Danilo D. Vaz
     */
    public function executeQuery (string $dataSourceName, array $filter = array(), string $query = null, array $columnTypes = array())
    {
        if ($dataSourceName) {
            $dataSource = $this->nameProvider->getDataSourceByName($dataSourceName);
            $query = $this->getQueryDataSource($dataSource);
            $columnTypes = $dataSource->getConditions();
        }
        
        $std = $this->connection->prepare($query);
        $std->execute($filter);

        $result = $std->fetchAll();
        
        $this->convertResultData($result, $columnTypes);
        
        return $result;
    }
    
    /**
     * Function that convert every element in array
     * 
     * Receives the query result and columns types and convert the values for
     * each element from array.
     * 
     * @param array $data Array that receives the query result
     * @param array $columnTypes Array that receives the columns types ['COLUMN'=>'TYPE']
     *
     * @return void
     * 
     * @access public
     * @author Danilo D. Vaz
     */
    private function convertResultData (&$data, $columnTypes)
    {
        foreach ($data as &$item) {
            foreach ($columnTypes as $column=>$type) {
                $item[$column] = $this->convertValue($type, $item[$column]);
            }
        }
    }
    
    /**
     * Function that convert a value
     * 
     * Convert the value according of type informed. 
     * The type can be: INT, FLOAT, BOOL, STRING and ARRAY.
     * 
     * @param string $type String that receives the type
     * @param string $value String that receives the value
     *
     * @return converted value
     * 
     * @access public
     * @author Danilo D. Vaz
     */
    private function convertValue ($type, $value)
    {
        switch ($type) {
            case 'INT':
                $converted = (int)$value;
                break;
            case 'FLOAT':
                $converted = (float)$value;
                break;
            case 'BOOL':
                $converted = (bool)$value;
                break;
            case 'STRING':
                $converted = (string)$value;
                break;
            case 'ARRAY':
                $converted = $value ? explode(';', $value) : $value;
                break;
            default:
                $converted = $value;
        }
        
        return $converted;
    }
    
    /**
     * Function that provides dataSource configuration
     * 
     * Process the Filter Criteria and returns a Configuration object with
     * datasource data
     * 
     * @param FilterCriteria $filterCriteria Object that receives filter data
     *
     * @return object Configuration
     * 
     * @access private
     * @author Danilo D. Vaz
     */
    private function getDataSource ($filterCriteria)
    {
        $dataSourceName = $filterCriteria->getDataSourceName();
        $dataSource = $this->nameProvider->getDataSourceByName($dataSourceName);
        
        if (!$dataSource) {
            throw new \Exception("DataSource '$dataSourceName' not found");
        }
        
        return $dataSource;
    }

    /**
     * Function that process the filter
     * 
     * Process the filter provides from request and the addtional filter, 
     * building a new filter with all information.
     * If informed filters with the same index, the value will be replaced.
     * 
     * @param FilterCriteria $filterCriteria Object that receives filter data
     * @param array $filterUri Associative array provides from uri ['COLUMN'=>'VALUE']
     * @param array $addtionalFilter Associative array with filter ['COLUMN'=>'VALUE']
     *
     * @return array
     * 
     * @access public
     * @author Danilo D. Vaz
     */
    public function processFilter (FilterCriteria $filterCriteria, $filterUri = array(), $addtionalFilter = array())
    {
        $conditions = $filterCriteria->getConditions();
        $filter = array();

        foreach ($conditions as $condition) {
            if (is_array($condition['value']) && isset($condition['value'][0])) {
                $value = implode('_', $condition['value']);
            } else {
                $value = $condition['value'];
            }
            
            $filter[strtoupper($condition['columnName'])] = $value;
        }
        
        $filterUri = array_change_key_case($filterUri, CASE_UPPER);
        $addtionalFilter = array_change_key_case($addtionalFilter, CASE_UPPER);

        return array_merge($filter, $filterUri, $addtionalFilter);
    }

    /**
     * Function that build the query
     * 
     * Process the data provides from Filter Criteria, like pagination,
     * search, range and ordination, and build the query adding row number and 
     * total amount of lines.
     * 
     * @param FilterCriteria $filterCriteria Object that receives filter data
     * @param Configuration $dataSource Object that receives datasource data
     * @param array $filter Array that receives the filter processed
     *
     * @return string
     * 
     * @access private
     * @author Danilo D. Vaz
     */
    private function buildQuery ($filterCriteria, $dataSource, &$filter)
    {
        $query = $this->getQueryDataSource($dataSource);
        $columns = $this->getColumnsDataSource($dataSource);
        $search = $this->processSearch($filterCriteria);
        $range = $this->processRange($filter, $search);
        $pagination = $this->processPagination($filterCriteria);
        $orderBy = $this->processOrderBy($filterCriteria);

        $finalQuery = "SELECT $columns FROM (
                            SELECT
                                USER_QUERY.*, 
                                ROWNUM RNUM, 
                                COUNT(1) OVER () TOTAL_LINES
                            FROM (
                                $query
                            ) USER_QUERY
                            $search
                            $range
                        ) $pagination
                        $orderBy";

        return strtoupper($finalQuery);
    }

    /**
     * Function that get query from datasource file
     * 
     * Receives datasource name and returns the query informed from datasource file.
     * 
     * @param Configuration $dataSource Object that receives datasource data
     *
     * @return string
     * 
     * @access private
     * @author Danilo D. Vaz
     */
    private function getQueryDataSource ($dataSource)
    {
        $query = $dataSource->getQuery();

        if (!$query) {
            $dataSourceName = $dataSource->getName();
            throw new \Exception("Query for data source '$dataSourceName' not found");
        }

        return $query;
    }

    /**
     * Function that get columns from datasource file
     * 
     * Receives datasource name and returns the columns informed from datasource file.
     * 
     * @param Configuration $dataSource Object that receives datasource data
     *
     * @return string
     * 
     * @access private
     * @author Danilo D. Vaz
     */
    private function getColumnsDataSource ($dataSource)
    {
        $columns = $dataSource->getColumns();
        
        if (count($columns)) {
            $columns = array_map(function($column) {
                return "\"$column\"";
            }, $columns);
            
            $columns = implode(', ', $columns) . ', TOTAL_LINES';
        } else {
            $dataSourceName = $dataSource->getName();
            throw new \Exception("Columns for data source '$dataSourceName' not found");
        }

        return $columns;
    }

    /**
     * Function that process search
     * 
     * Receives a Filter Criteria object and build the where clause according 
     * to the search. Returns where clause that will be adding in main query.
     * 
     * @param FilterCriteria $filterCriteria Object that receives filter data
     *
     * @return string
     * 
     * @access private
     * @author Danilo D. Vaz
     */
    private function processSearch ($filterCriteria)
    {
        $search = $filterCriteria->getSearch();
        $searchIn = $filterCriteria->getSearchIn();

        if (count($search) && count($searchIn)) {
            $value = '%' . implode('%', $search) . '%';
            $where = 'WHERE ';
            $lastKey = array_key_last($searchIn);

            foreach ($searchIn as $index=>$column) {
                $where .= "(UPPER(\"$column\") LIKE UPPER('$value'))";

                if ($lastKey != $index) {
                    $where .= " OR ";
                }
            }
        } else if (!count($search) && !count($searchIn)) {
            $where = null;
        } else {
            throw new \Exception("The 'search' or 'search_in' not found in uri");
        }

        return $where;
    }
    
    /**
     * Function that process range
     * 
     * Receives the filter processed and build the where clause according 
     * to the tags 'greater_than_equals' and 'less_than_equals', removing from 
     * filter this values. Checks the search clause for build the range clause
     * together. Returns where clause that will be added in main query.
     * 
     * @param array $filter Array that receives filter data
     * @param string $searchClause String that receives search clause
     *
     * @return string
     * 
     * @access private
     * @author Danilo D. Vaz
     */
    private function processRange (&$filter, $searchClause)
    {
        $range = array();
        $where = null;
        
        foreach ($filter as $column=>$value) {
            if (is_array($value)) {
                if (array_key_exists('greater_than_equals', $value) || array_key_exists('less_than_equals', $value)) {
                    if (isset($value['greater_than_equals'])) {
                        $greaterThan = $value['greater_than_equals'];
                        array_push($range, "\"$column\" >= $greaterThan");
                    }
                    
                    if (isset($value['less_than_equals'])) {
                        $lessThan = $value['less_than_equals'];
                        array_push($range, "\"$column\" <= $lessThan");
                    }
                    
                    unset($filter[$column]);
                }
            }
        }
        
        if (count($range)) {
            $where = $searchClause ? 'AND ' : 'WHERE ';
            $where .= implode(' AND ', $range);
        }

        return $where;
    }

    /**
     * Function that process pagination
     * 
     * Receives a Filter Criteria object and build the where clause according 
     * to the pagination. Returns where clause that will be adding in main query.
     * If not informed pagination, the page size default of framework is 300.
     * 
     * @param FilterCriteria $filterCriteria Object that receives filter data
     *
     * @return string
     * 
     * @access private
     * @author Danilo D. Vaz
     */
    private function processPagination ($filterCriteria)
    {
        $page = $filterCriteria->getPage();
        $pageSize = $filterCriteria->getPageSize();

        if ($page) {
            $end = $page * $pageSize;
            $init = $end - $pageSize + 1;
        } else {
            $filterCriteria->setPage(1);
            $end = $pageSize;
            $init = 1;
        }

        $where = "WHERE RNUM BETWEEN $init AND $end";

        return $where;
    }

    /**
     * Function that process ordination
     * 
     * Receives a Filter Criteria object and build the where clause according 
     * to the ordination. Returns where clause that will be adding in main query.
     * 
     * @param FilterCriteria $filterCriteria Object that receives filter data
     *
     * @return string
     * 
     * @access private
     * @author Danilo D. Vaz
     */
    private function processOrderBy ($filterCriteria)
    {
        $orderBy = $filterCriteria->getOrderBy();
        $where = null;

        if (count($orderBy)) {
            $where = "ORDER BY ";
            $lastKey = array_key_last($orderBy);

            foreach ($orderBy as $column=>$direction) {
                $where .= "\"$column\" $direction";

                if ($lastKey != $column) {
                    $where .= ", ";
                }
            }
        }

        return $where;
    }
    
    /**
     * Function that build uri
     * 
     * Receives and modify the query result, adding the element uri.
     * Can receives an array with the primary keys or an object with datasource
     * configuration.
     * 
     * @param array $result Array with query result
     * @param array $primaryKeys Array with primary keys ['NAME', ...]
     * @param array $filterUri Array that define if url have already the PK
     * @param Configuration $dataSource Object that receives datasource data
     *
     * @return void
     * 
     * @access public
     * @author Danilo D. Vaz
     */
    public function buildUri (array &$result, array $primaryKeys, $filterUri = false, $dataSource = null)
    {
        if ($dataSource instanceof Configuration) {
            $primaryKeys = $dataSource->getPrimaryKeyColumns();
        }
        
        $requestScheme = $_SERVER['REQUEST_SCHEME'];
        $host = $_SERVER['HTTP_HOST'];
        $path = $_SERVER['PHP_SELF'];
        
        $productPath = explode('/', $this->productPath);
        $path = explode('/', ltrim($path, '/'));

        $fullPath = implode('/', array_unique(array_merge($productPath, $path)));
        
        $uri = $requestScheme . '://' . $host . '/' . $fullPath;

        foreach ($result as &$item) {
            $item['uri'] = $uri;
            
            if (empty($filterUri)) {
                foreach ($primaryKeys as $primaryKey) {
                    $item['uri'] .= '/' . $item[$primaryKey];
                }
            }
        }
    }

    /**
     * Function that build the data
     * 
     * Receives the query result and the Filter Criteria, and build a data array
     * in the default expected by the frontend.
     * 
     * @param array $result Array with query result
     * @param FilterCriteria $filterCriteria Object that receives filter data
     *
     * @return array
     * 
     * @access public
     * @author Danilo D. Vaz
     */
    public function buildData (array $result, FilterCriteria $filterCriteria)
    {
        $page = $filterCriteria->getPage();
        $pageSize = $filterCriteria->getPageSize();
        
        if (isset($result[0]['TOTAL_LINES'])) {
            $total = $result[0]['TOTAL_LINES'];
        } else if (isset($result['TOTAL_LINES'])) {
            $total = $result['TOTAL_LINES'];
        } else {
            $total = count($result);
        }

        $pagination = array(
            "total" => (int)$total,
            "limit" => $pageSize,
            "page" => $page
        );

        $data = array(
            'data' => $result,
            'pagination' => $pagination
        );

        return $data;
    }

    /**
     * Function to set connection
     * 
     * Function that set a new Connection to be used by CRUD functions.
     * 
     * @param Connection $connection Object that receives a new connection
     *
     * @return void
     * 
     * @access public
     * @author Pedro Henrique Vilaça Valverde
     */
    public function setConnection (Connection $connection){
        $this->connection = $connection;
    }

}