<?php

namespace SupplyTek\Utilities;

use SupplyTek\Utilities\SMTP;
use SupplyTek\Utilities\SessionHandler;
use SupplyTek\Utilities\TemplateHandler;
use SupplyTek\Utilities\CrudSupplyTek;
use Zeedhi\Framework\DependencyInjection\InstanceManager;

class EmailHandler
{
    protected $session;
    protected $template;
    protected $crud;
    protected $fileServerURL;
    protected $smtp;
    
    private $useTemplate = true;
    private $isUsingCustomSmtp = false;
    private $customSmtpSettings = array();
    
    public function __construct(SessionHandler $sessionHandler, TemplateHandler $templateHandler, CrudSupplyTek $crudSupplyTek, $fileServerURL) 
    {
        $this->session = $sessionHandler;
        $this->template = $templateHandler;
        $this->crud = $crudSupplyTek;
        $this->fileServerURL = $fileServerURL;
    }
    
    public function send($subject, $content, $emails)
    {
        $this->useCustomSmtp();
        
        if ($this->useTemplate) {
            $unitId = $this->session->getSessionData("CDFILIAL");
            
            $logo = $this->getLogo();
            $lgpd = $this->getLgpd($unitId);
            
            $imgAttr = "width='175px'";
            $this->getLogoAttributes($logo, $imgAttr);
            
            $contentHeader = "<p class='bold' style='text-align: center; color: #ED2424'>" . $subject . "</p>"; 
            
            $file = array("name" => "email", "extension" => ".html");
            $replace = array(
                "%ATTRIMG%" => $imgAttr, 
                "%LOGOSRC%" => $logo, 
                "%SUBJECT%" => $contentHeader, 
                "%MESSAGE%" => $content, 
                "%LGPD%" => $lgpd
            );
            
            $content = $this->template->getTemplate($file, $replace);
        } 
        
        $this->smtp->setSubject($subject);
        $this->smtp->sendMail($content, $emails);
    }
    
    public function useTemplate($bool)
    {
        $this->useTemplate = $bool;
    }
    
    private function getLgpd($unitId)
    {
        $params = array(
            "UNIT_ID" => $unitId
        );
        
        $result = $this->crud->executeQuery("lgpd-email", $params);
        
        return isset($result[0]) ? $result[0]['LGPD'] : "";
    }
    
    private function getLogo($appId = 'PORTAL E-COMMERCE') 
    {
        $params = array(
            "APP_ID" => $appId,
            "ORG_ID" => $this->session->getSessionData("NRORG")
        );
        
        $result = $this->crud->executeQuery("logo-email", $params);
        
		if (isset($result[0]['LOGO'])) {
    		return $this->fileServerURL . $result[0]['LOGO'];
        } else {
            return false;
        }
    }
    
    private function getLogoAttributes(&$logoSrc, &$attrImg = 'width="200px"')
    {
        if ($logoSrc != false && $this->checkFileExists($logoSrc)) {
            $imgAttributes = getimagesize($logoSrc);
            
            if ($imgAttributes[1] > $imgAttributes[0]) {
                $attrImg = 'height="175px"';
            }
        } else {
            $logoSrc = 'https://www.teknisa.com/imagens/logo-teknisa.png';
        }
    }
    
    private function checkFileExists($url)
    {
        $ch = curl_init($url);
        curl_setopt($ch, CURLOPT_NOBODY, true);
        curl_exec($ch);
        $code = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        
        if ($code == 200) {
            $status = true;
        } else {
            $status = false;
        }
        
        curl_close($ch);
        return $status;
    }
    
    /*********************************************************************************************/
    /**************************************** CUSTOM SMTP ****************************************/
    /*********************************************************************************************/
    
    public function useCustomSmtp()
    {
        if (!$this->isUsingCustomSmtp) {
            $this->customSmtpSettings = $this->getCustomSmtpSettings();
        }
    
        $this->smtp = new SMTP($this->customSmtpSettings);

        if (isset($this->customSmtpSettings['fromName'])) {
            $this->smtp->setFromName($this->customSmtpSettings['fromName']);
        }

        if (isset($this->customSmtpSettings['from'])) {
            $this->smtp->setNameReplyTo($this->customSmtpSettings['from']);
        }
    
        $this->isUsingCustomSmtp = true;
    }

    public function getCustomSmtpSettings()
    {
        $tipoDadoUtilizado = $this->getTipoDadoUilizadoEnvioEmail()['DATA_TYPE'];
        $dadosSMTPFilial   = $this->getDadosSMTPFromFilial();
        $settings          = self::getSmtpSettingsFromAddress($dadosSMTPFilial['SMTP_EMAIL']);

        $params['host'] = $settings['host'];
        
        if (isset($settings['port'])) {
            $params['port'] = $settings['port'];
        }

        switch($tipoDadoUtilizado) {
            case 'LOGIN_FILIAL_REMETENTE_FILIAL':
                $params['from']     = $dadosSMTPFilial['FROM_NAME'];
                $params['fromName'] = $dadosSMTPFilial['FROM_NAME'];
                $params['username'] = $dadosSMTPFilial['USERNAME'];
                $params['password'] = $dadosSMTPFilial['PASSWORD'];
            break;
             
            case 'LOGIN_FILIAL_REMETENTE_OPER':
                $dadosEmailOperadorLogado = $this->getDadosEmailOperadorLogado();
                
                $params['from']     = $dadosEmailOperadorLogado['USERNAME'];
                $params['fromName'] = $dadosEmailOperadorLogado['FROM_NAME'];
                $params['username'] = $dadosSMTPFilial['USERNAME'];
                $params['password'] = $dadosSMTPFilial['PASSWORD'];
            break;
            
            case 'LOGIN_OPER_REMETENTE_OPER':
                $dadosEmailOperadorLogado = $this->getDadosEmailOperadorLogado();

                $params['from']     = $dadosEmailOperadorLogado['USERNAME'];
                $params['fromName'] = $dadosEmailOperadorLogado['FROM_NAME'];
                $params['username'] = $dadosEmailOperadorLogado['USERNAME'];
                $params['password'] = $dadosEmailOperadorLogado['PASSWORD'];
            break;
        }
        
        if (!empty($params['password'])) {
            $params['password'] = self::getDelphiPasswordEncrypted($params['password'], false);
        }

        $securityParameters = $this->getSecurityParameters($dadosSMTPFilial, $params);
        $params = array_merge($params, $securityParameters);

        return $params;
    }
    
    /**
     * Recupera informações do endereço de SMTP, como:
     * - host;
     * - porta;
     * 
     * @param string $url Url do servidor com ou sem porta.
     * 
     * @return array Contendo as informações encontradas.
     */
    public function getSmtpSettingsFromAddress($url)
    {
        $match = self::splitByCommonDelimiters($url);
        
        if (empty($match)) {
            throw new \Exception('Formato do endereço de SMTP inválido.');
        }
        
        if (count($match) > 1) {
            $params['port'] = $match[1];
        }
        
        $params['host'] = $match[0];
        
        return $params;
    }
    
    public function splitByCommonDelimiters($string)
    {
        $pattern = '/[:;|]+/';
        $cleanedString = self::removeAllWhitespaces($string);
        
        $splittedString = preg_split($pattern, $cleanedString);
        
        return $splittedString;
    }
    
    public function removeAllWhitespaces($string)
    {
        $pattern = '/\s+/';
        
        $cleanedString = preg_replace($pattern, '', $string);
        
        return $cleanedString;
    }
    
    public function getTipoDadoUilizadoEnvioEmail()
    {
        $buyerUnitId = self::getCentralCompras($this->session->getSessionData("CDFILIAL"))['BUYER_UNIT_ID'];
        
        $params = array(
            "BUYER_UNIT_ID" => $buyerUnitId
        );
        
        $result = $this->crud->executeQuery("data-type-email", $params);
        
        return isset($result[0]) ? $result[0] : $result;
    }
    
    public function getCentralCompras($unitId)
    {
        $params = array(
            "UNIT_ID" => $unitId
        );
        
        $result = $this->crud->executeQuery("buyer-unit-email", $params);
        
        return isset($result[0]) ? $result[0] : $result;
    }
    
    public function getDadosSMTPFromFilial()
    {
        $params = array(
            "UNIT_ID" => $this->session->getSessionData("CDFILIAL")
        );
        
        $result = $this->crud->executeQuery("unit-smtp-data-email", $params);

        return isset($result[0]) ? $result[0] : $result;
    }
    
    public function getDadosEmailOperadorLogado()
    {
        $params = array(
            "OPERATOR_ID" => $this->session->getSessionData("CDOPERADOR")
        );
        
        $result = $this->crud->executeQuery("logged-operator-data-email", $params);

        return isset($result[0]) ? $result[0] : $result;
    }
    
    /**
     * Temporário até que uma solução seja implementada no local original desta
     * função.
     * 
     * A original realiza chamada apenas nas funções GetCriptoPass e GetUnCriptoPass,
     * equivalentes ao CriptoPlus e UncriptoPlus do Delphi.
     * 
     * Porém, utilizamos as criptografias convencionais, que até então não eram
     * implementadas la Bib_html.
     * 
     * Foi incluída na 'Bib', rdm 'rdmDBGeral', na issue #238324
     * 
     * Teknisa\Libs\Util\Utilities
     */
    public static function getDelphiPasswordEncrypted($password, $isEncrypt = true, $offset = '2')
    {
        $url = self::getAspUrlParameter() . '/comwebservice/service.asmx?wsdl';
        
        $options = array(
            'encoding'    => 'ISO-8859-1',
            'trace'       => true,
            'exceptions'  => true,
            'compression' => SOAP_COMPRESSION_ACCEPT | SOAP_COMPRESSION_GZIP,
        );
        
        $args = array(
            new \SoapVar($password, XSD_STRING, 'string', "http://www.w3.org/2001/XMLSchema"),
            new \SoapVar($offset, XSD_STRING, 'string', "http://www.w3.org/2001/XMLSchema"),
        );
        
        $cryptType = $isEncrypt ? 'GetCripto' : 'GetUncripto';
        
        $params = array(
            'rdm'    => 'Bib.rdmDBGeral',
            'method' => $cryptType,
            'args'   => $args,
        );
        
        try {
            $client = new \SoapClient($url, $options);
            $result = $client->executeCOMMethod($params);
            $encryption = $result->executeCOMMethodResult;
        } catch (\SoapFault $e) {
            throw new \Exception($e->getMessage());
        }

        return $encryption;
    }
    
    /**
     * Temporário até que uma solução seja implementada para a função getDelphiPasswordEncrypted.
     * 
     * Teknisa\Libs\Util\Utilities
     */
    public static function getAspUrlParameter() {
        return InstanceManager::getInstance()->getParameter('aspUrl');
    }

    /**
     * Recupera informações de segurança para usar na conexão com o SMTP.
     * Utiliza dos parâmetros recuperados para definir as seguintes propriedades:
     * . smtpSecure:
     *      - 'tls', caso o parâmetro IDUTTLSEMAIL esteja marcado;
     *      - 'ssl', caso o parâmetro IDUTSSLEMAIL esteja marcado;
     *      - '', caso ambos os parâmetros acima estejam desmarcados;
     * . smtpAuth:
     *      - true, caso o usuário e senha estejam preenchidos;
     *      - false, caso usuário ou senha estejam em branco;
     * 
     * @param array $dadosSMTPFilial
     * @param array $dadosSMTPCustom
     * 
     * @return array Contendo os parâmetros de segurança.
     */
    private function getSecurityParameters($dadosSMTPFilial, $dadosSMTPCustom)
    {
        $params = [
            'smtpSecure' => '',
            'smtpAuth' => false
        ];
        
        if ($dadosSMTPFilial['USE_TLS_EMAIL'] == 'S') {
            $params['smtpSecure'] = 'tls';
        } else if ($dadosSMTPFilial['USE_SSL_EMAIL'] == 'S') {
            $params['smtpSecure'] = 'ssl';
        }
        
        if (!empty($dadosSMTPCustom['username']) && !empty($dadosSMTPCustom['password'])) {
            $params['smtpAuth'] = true;
        }
        
        return $params;
    }
} 